import hjson
import os
import copy
from typing import Callable, List, Dict, Union
from dlk.utils.register import Register
from dlk.utils.config import ConfigTool
from dlk.utils.logger import Logger
from dlk.utils.get_root import get_root
import json

logger = Logger.get_logger()

config_parser_register = Register("Config parser register")


class LinkUnionTool(object):
    """
       Assisting tool for parsering the "_link" of config
    """
    def __init__(self):
        self.link_union = {}

    def find(self, key):
        if key not in self.link_union:
            return None

        if self.link_union[key] != key:
            return self.find(self.link_union[key])
        return self.link_union[key]

    def low_level_union(self, parant, child):
        """
            On the basis of the high-level links, `low_level_union` should be used when regist low-level link
            If parent and child wer all not appeared at before, we will directly regist them.
            If only one of the parent and child appeared, the value of the parent and child will be overwritten by the corresponding value of the upper level,
            If they both appeared at before, and if they linked the same value, we will do nothing, otherwise `RAISE AN ERROR`
        """
        if self.find(parant) and self.find(child):  # all has been linked
            if self.find(parant)!= self.find(child):
                raise PermissionError(f"The  {parant} and {child} has been linked to different values, but now you want to link them together.")
            elif self.find(parant) == child:
                logger.warning(f"High level config has the link '{child} -> {parant}', and the low level reversed link '{parant} -> {child}' is been ignored.")
            else:
                return
        elif self.find(child): # only child has been linked
            logger.warning(f"Parameter '{child}' has been linked in high level config, the link '{parant} -> {child}' is invalid, and the real link is been reversed as '{child} -> {parant}'.")
            self.link_union[parant] = self.find(child)
        elif self.find(parant): # only parant has been linked
            self.link_union[child] = self.find(parant)
        else:
            self.link_union[parant] = parant
            self.link_union[child] = parant

    def top_level_union(self, parant, child):
        """
            Register the 'link'(parant -> child) in the same(top) level config should be merged using `same_level_union`

            Parameters are not allowed to be assigned repeatedly (the same parameter cannot appear more than once in the target position, otherwise it will cause ambiguity.)
        """
        if parant not in self.link_union:
            self.link_union[parant] = parant

        assert child not in self.link_union, f"{child} is repeated assignment"
        self.link_union[child] = self.find(parant)

    def register_top_links(self, links: Dict):
        """regist the top level links
        :links: dict: {"from": ["tolist"], "from2": "to2"}
        :returns: self
        """
        for source, target in links.items():
            if isinstance(target, list):
                [self.top_level_union(source, t) for t in target]
            else:
                self.top_level_union(source, target)
        return self

    def register_low_links(self, links: Dict):
        """regist the low level links
        :links: dict: {"from": ["tolist"], "from2": "to2"}
        :returns: self
        """
        for source, target in links.items():
            if isinstance(target, list):
                [self.low_level_union(source, t) for t in target]
            else:
                self.low_level_union(source, target)
        return self

    def get_links(self):
        """
            return the registed links as a dict
        """
        links = {}
        for key in self.link_union:
            parant = self.find(key)
            if parant == key:
                continue
            if parant not in links:
                links[parant] = []
            links[parant].append(key)
        return links


class BaseConfigParser(object):
    """docstring for BaseConfigParser

        input: config_file, config_base_dir
        config_name = config_file._name
        search = config_file._search
        link = config_file._link
        base = self.config_file._base
        if base:
            base_config = parser(base)
    """
    def __init__(self, config_file: Union[str, Dict, List], config_base_dir: str=""):
        super(BaseConfigParser, self).__init__()
        if isinstance(config_file, str):
            if config_file == '*@*':
                self.config_file = "*@*"
                return
            try:
                self.config_file = self.load_hjson_file(os.path.join(get_root(), config_base_dir, config_file+'.hjson'))
            except Exception as e:
                logger.error(f"There is an error occur when loading {os.path.join(get_root(), config_base_dir, config_file)}")
                raise KeyError(e)
        elif isinstance(config_file, Dict):
            self.config_file = config_file
        else:
            raise KeyError('The config file must be str or dict. You provide {}.'.format(config_file))

        self.config_name = self.config_file.pop('_name', "")
        self.search = self.config_file.pop('_search', {})
        base = self.config_file.pop('_base', "")
        self.base_config = {}
        if base:
            self.base_config = self.get_base_config(base)
        if "_focus" in self.config_file:
            self.base_config['_focus'] = self.config_file.pop('_focus')

        # merge base and current config _link
        link_union = LinkUnionTool()
        link_union.register_top_links(self.config_file.pop('_link', {}))
        link_union.register_low_links(self.base_config.pop('_link', {}))
        self.config_file['_link'] = link_union.get_links()

        if self.base_config and self.config_name:
            raise PermissionError("You should put the _name to the leaf config.")
        self.modules = self.config_file


    @classmethod
    def get_base_config(cls, config_file)->Dict:
        """TODO: Docstring for get_base_config.

        :arg1: TODO
        :returns: TODO

        """
        base_config = cls(config_file).parser(parser_link=False)
        if len(base_config)>1:
            raise PermissionError("The base config don't support _search now.")
        if base_config:
            return base_config[0]
        return {}

    @staticmethod
    def config_link_para(link: Dict[str, Union[str, List[str]]]={}, config: Dict={}):
        """inplace link the config[to] = config[source]
        :link: {source1:to1, source2:[to2, to3]}
        :returns:
        """
        def make_link(source: str, to: str):
            """copy the 'source' config to 'to'
            """
            try:
                source_config = config
                to_config = config
                source_list = source.split('.')
                to_list = to.split('.')
                for s in source_list[:-1]:
                    if isinstance(source_config, list):
                        assert (s[0]=='-' and str.isdigit(s[1:])) or str.isdigit(s), "list index must be int"
                        s = int(s)
                    source_config = source_config[s]
                for t in to_list[:-1]:
                    if isinstance(to_config, list):
                        assert (t[0]=='-' and str.isdigit(t[1:])) or str.isdigit(t), "list index must be int"
                        t = int(t)
                    to_config = to_config[t]
                if isinstance(to_config, list):
                    assert (to_list[-1][0]=='-' and str.isdigit(to_list[-1][1:])) or str.isdigit(to_list[-1]), "list index must be int"
                    to_list[-1] = int(to_list[-1])

                if isinstance(source_config, list):
                    assert (source_list[-1][0]=='-' and str.isdigit(source_list[-1][1:])) or str.isdigit(source_list[-1]), "list index must be int"
                    source_list[-1] = int(source_list[-1])

                to_config[to_list[-1]] = source_config[source_list[-1]]
            except Exception as e:
                logger.error(f"Can not link from '{source}' to '{to}'")
                raise e


        if not link:
            return
        for (source, to) in link.items():
            if isinstance(to, List):
                for sub_to in to:
                    make_link(source, sub_to)
            else:
                make_link(source, to)

    @classmethod
    def collect_link(cls, config, trace=[], all_level_links={}, level=0):
        """
            only do in the top level of config, collect all level links and return the links with level
        """
        if level not in all_level_links:
            all_level_links[level] = {}
        trace_str = ".".join(trace)
        if trace_str:
            # except the top level, all add the '.' suffix
            trace_str = trace_str + '.'
        def add_trace(origin_link: Dict)->Dict:
            """add the root of the config to current config trace to current level para of links
            :returns: the origin_link added trace

            """
            added_trace_link = {}
            for source, target in origin_link.items():
                if isinstance(target, list):
                    target = [trace_str+t for t in target]
                else:
                    target = trace_str + target
                source = trace_str + source
                added_trace_link[source] = target
            return added_trace_link

        if "_link" in config:
            # all_level_links[level] = add_trace(config['_link'])
            all_level_links[level].update(add_trace(config.pop('_link')))

        for submodule_name, submodule_config in config.items():
            if isinstance(submodule_config, dict):
                cls.collect_link(submodule_config, trace + [submodule_name], all_level_links, level+1)
        return all_level_links

    def parser_with_check(self, parser_link=True)->List[Dict]:
        """parser config and check the result, only used for __main__ processor
        :returns: parserd all possible configs
        """
        configs = self.parser(parser_link)
        self.check_config(configs)
        return configs

    def parser(self, parser_link=True) -> List:
        """ root parser
        return a list of para dicts for all possible combinations
        :parser_link: whether parser the link of config
        :returns: list[possible config dict]
        """
        if self.config_file == '*@*':
            return ['*@*']

        # parser submodules get submodules config
        modules_config = self.map_to_submodule(self.modules, self.get_kind_module_base_config)

        # expand all submodules to combine a set of module configs
        possible_config_list = self.get_named_list_cartesian_prod(modules_config)

        # using specifical module config to update base_config
        if possible_config_list:
            possible_config_list = [ConfigTool.do_update_config(self.base_config, possible_config) for possible_config in possible_config_list]
        else:
            possible_config_list = [self.base_config]

        # flat all search paras
        possible_config_list_list = [self.flat_search(self.search, possible_config) for possible_config in possible_config_list]

        all_possible_config_list = []
        for possible_config_list in possible_config_list_list:
            all_possible_config_list.extend(possible_config_list)

        # link paras
        if parser_link:
            for possible_config in all_possible_config_list:
                all_level_links = self.collect_link(possible_config)
                link_union = LinkUnionTool()
                for i in range(len(all_level_links)):
                    cur_level_links = all_level_links[i]
                    link_union.register_low_links(cur_level_links)

                self.config_link_para(link_union.get_links(), possible_config)

        return_list = []
        for possible_config in all_possible_config_list:
            config = copy.deepcopy(possible_config)
            if self.config_name:
                config['_name'] = self.config_name
            return_list.append(config)

        if self.is_rep_config(return_list):
            logger.warning(f"The Configures is Repeated, Please Check The Configures Carefully.")
            for i, config in enumerate(return_list):
                logger.info(f"The {i}th Configure is:")
                logger.info(json.dumps(config, indent=2))
            raise ValueError('REPEAT CONFIG')
        return return_list

    def get_kind_module_base_config(self, abstract_config: Union[dict, str], kind_module: str="") -> List[dict]:
        """get the whole config of 'kind_module' by given abstract_config

        :abstract_config: dict: will expanded config
        :returns: parserd config (whole config) of abstract_config

        """
        return config_parser_register.get(kind_module)(abstract_config).parser(parser_link=False)

    def map_to_submodule(self, config: dict, map_fun: Callable) -> Dict:
        """use the map_fun to process all the modules
        :returns: depend on

        """
        modules_config = {}
        for kind_module in config:
            modules_config[kind_module] = map_fun(config[kind_module], kind_module)
        return modules_config

    def load_hjson_file(self, file_name: str) -> Dict:
        """load hjson file by file_name

        :file_name: TODO
        :returns: TODO

        """
        json_file = hjson.load(open(file_name), object_pairs_hook=dict)
        return json_file

    # def flat_search(self, config: dict) -> List[dict]:
    @classmethod
    def flat_search(cls, search, config: dict) -> List[dict]:
        """flat all the _search paras to list
        support recursive parser _search now, this means you can add _search/_link/_base paras in _search paras

        :config: dict: base config
        :returns: list of possible config

        """
        result = []
        module_search_para = search
        if not module_search_para:
            result.append(config)
        else:
            search_para_list = cls.get_named_list_cartesian_prod(module_search_para)
            for search_para in search_para_list:
                base_config = copy.deepcopy(config)
                base_config.update(search_para)
                extend_config = cls(base_config).parser(parser_link=False)
                result.extend(extend_config)
                # result.append(base_config)

        return result

    def get_cartesian_prod(self, list_of_list_of_dict: List[List[Dict]]) -> List[List[Dict]]:
        """get catesian prod from two lists
        :list_of_list_of_dict: [[config_a1, config_a2], [config_b1, config_b2]]
        :returns: [[config_a1, config_b1],
                   [config_a1, config_b2],
                   [config_a2, config_b1],
                   [config_a2, config_b2]
                  ]
        """
        if len(list_of_list_of_dict) <= 1:
            return [copy.deepcopy(dic) for dic in list_of_list_of_dict]
        cur_result = list_of_list_of_dict[0]
        reserve_result = self.get_cartesian_prod(list_of_list_of_dict[1:])
        result = []
        for cur_config in cur_result:
            for reserve in reserve_result:
                result.append([copy.deepcopy(cur_config)]+copy.deepcopy(reserve))
        return result

    @staticmethod
    def check_config(configs: Union[Dict, List[Dict]]) -> None:
        """check all config is right.
        1. check all "*@*" is replaced to correct value.
        :configs: one config or a list of configs
        """
        def _check(config):
            """TODO: Docstring for _check.

            :config: TODO
            :returns: TODO

            """
            for key in config:
                if isinstance(config[key], dict):
                    _check(config[key])
                if config[key] == '*@*':
                    raise ValueError(f'In Config: \n {json.dumps(config, indent=4)}\n The must be provided key "{key}" marked with "*@*" is not provided.')

        if isinstance(configs, list):
            for config in configs:
                _check(config)
        else:
            _check(configs)

    @staticmethod
    def get_named_list_cartesian_prod(dict_of_list: Dict[str, List]={}) -> List[Dict]:
        """get catesian prod from named lists
        :dict_of_list: {'name1': [1,2,3], 'name2': [1,2,3]}
        :returns: [{'name1': 1, 'name2': 1}, {'name1': 1, 'name2': 2}, {'name1': 1, 'name2': 3}, ...]

        """
        if len(dict_of_list) == 0:
            return []
        dict_of_list = copy.deepcopy(dict_of_list)
        cur_name, cur_paras  = dict_of_list.popitem()
        cur_para_search_list = []
        if isinstance(cur_paras, str):
            cur_paras = eval(cur_paras)
        assert isinstance(cur_paras, list), f"The search options must be list, but you provide {cur_paras}({type(cur_paras)})"
        for para in cur_paras:
            cur_para_search_list.append({cur_name: para})
        if len(dict_of_list) == 0:
            return cur_para_search_list
        reserve_para_list = BaseConfigParser.get_named_list_cartesian_prod(dict_of_list)
        all_config_list = []
        for cur_config in cur_para_search_list:
            for reserve_config in reserve_para_list:
                _cur_config = copy.deepcopy(cur_config)
                _cur_config.update(copy.deepcopy(reserve_config))
                all_config_list.append(_cur_config)
        return all_config_list

    def is_rep_config(self, list_of_dict: List[dict]) -> bool:
        """check is there a repeat config in list

        :list_of_dict: List[dict]: TODO
        :returns: is there the list of dict repeat

        """
        # using json.dumps + sort_keys to guarantee the same dict to the same string represatation
        list_of_str = [json.dumps(dic, sort_keys=True) for dic in list_of_dict]
        if len(list_of_dict) == len(set(list_of_str)):
            return False
        else:
            return True


@config_parser_register('config')
class ConfigConfigParser(BaseConfigParser):
    """docstring for ConfigConfigParser"""
    def __init__(self, config_file):
        super(ConfigConfigParser, self).__init__(config_file, config_base_dir='NONEPATH')
        if self.base_config:
            raise AttributeError('The paras config do not support _base.')
        # if self.link:
            # raise AttributeError('The paras config do not support _link.')
        if self.config_name:
            raise AttributeError('The paras config do not support _name.')

    def parser(self, parser_link=True):
        """config support _search and _link

        :update_config: Dict: TODO
        :returns: TODO
        """
        config_list = self.flat_search(self.search, self.modules)
        # link paras
        if parser_link:
            for possible_config in config_list:
                all_level_links = self.collect_link(possible_config)
                link_union = LinkUnionTool()
                for i in range(len(all_level_links)):
                    cur_level_links = all_level_links[i]
                    link_union.register_low_links(cur_level_links)
                self.config_link_para(all_level_links, possible_config)

        return config_list


@config_parser_register('_link')
class LinkConfigParser(object):
    """docstring for LinkConfigParser"""
    def __init__(self, config_file):
        self.config = config_file
        assert isinstance(self.config, dict), f"The '_link' must be a dict, but you provide '{self.config}'"
    def parser(self, parser_link=False):
        """TODO: Docstring for parser.
        :parser_link: TODO
        :returns: TODO
        """
        assert parser_link is False, f"The parser_link para must be False when parser the _link"
        return [self.config]


@config_parser_register('task')
class TaskConfigParser(BaseConfigParser):
    """docstring for TaskConfigParser"""
    def __init__(self, config_file):
        super(TaskConfigParser, self).__init__(config_file, config_base_dir='dlk/configures/tasks/')


@config_parser_register('root')
class RootConfigParser(BaseConfigParser):
    """docstring for RootConfigParser"""
    def __init__(self, config_file):
        super(RootConfigParser, self).__init__(config_file, config_base_dir='')


@config_parser_register('manager')
class ManagerConfigParser(BaseConfigParser):
    """docstring for ManagerConfigParser"""
    def __init__(self, config_file):
        super(ManagerConfigParser, self).__init__(config_file, config_base_dir='dlk/configures/managers/')


@config_parser_register('callback')
class CallbackConfigParser(BaseConfigParser):
    """docstring for CallbackConfigParser"""
    def __init__(self, config_file):
        super(CallbackConfigParser, self).__init__(config_file, config_base_dir='dlk/configures/core/callbacks/')


@config_parser_register('datamodule')
class DatamoduleConfigParser(BaseConfigParser):
    """docstring for DatamoduleConfigParser"""
    def __init__(self, config_file):
        super(DatamoduleConfigParser, self).__init__(config_file, config_base_dir='dlk/configures/data/datamodules/')


@config_parser_register('imodel')
class IModelConfigParser(BaseConfigParser):
    """docstring for IModelConfigParser"""
    def __init__(self, config_file):
        super(IModelConfigParser, self).__init__(config_file, config_base_dir='dlk/configures/core/imodels/')


@config_parser_register('model')
class ModelConfigParser(BaseConfigParser):
    """docstring for ModelConfigParser"""
    def __init__(self, config_file):
        super(ModelConfigParser, self).__init__(config_file, config_base_dir='dlk/configures/core/models/')


@config_parser_register('optimizer')
class OptimizerConfigParser(BaseConfigParser):
    """docstring for OptimizerConfigParser"""
    def __init__(self, config_file):
        super(OptimizerConfigParser, self).__init__(config_file, config_base_dir='dlk/configures/core/optimizers/')


@config_parser_register('scheduler')
class ScheduleConfigParser(BaseConfigParser):
    """docstring for ScheduleConfigParser"""
    def __init__(self, config_file):
        super(ScheduleConfigParser, self).__init__(config_file, config_base_dir='dlk/configures/core/schedulers/')

@config_parser_register('initmethod')
class InitMethodConfigParser(BaseConfigParser):
    """docstring for InitMethodConfigParser"""
    def __init__(self, config_file):
        super(InitMethodConfigParser, self).__init__(config_file, config_base_dir='dlk/configures/core/initmethods/')


@config_parser_register('loss')
class LossConfigParser(BaseConfigParser):
    """docstring for LossConfigParser"""
    def __init__(self, config_file):
        super(LossConfigParser, self).__init__(config_file, config_base_dir='dlk/configures/core/losses/')


@config_parser_register('encoder')
class EncoderConfigParser(BaseConfigParser):
    """docstring for EncoderConfigParser"""
    def __init__(self, config_file):
        super(EncoderConfigParser, self).__init__(config_file, config_base_dir='dlk/configures/core/layers/encoders/')


@config_parser_register('decoder')
class DecoderConfigParser(BaseConfigParser):
    """docstring for DecoderConfigParser"""
    def __init__(self, config_file):
        super(DecoderConfigParser, self).__init__(config_file, config_base_dir='dlk/configures/core/layers/decoders/')


@config_parser_register('embedding')
class EmbeddingConfigParser(BaseConfigParser):
    """docstring for EmbeddingConfigParser"""
    def __init__(self, config_file):
        super(EmbeddingConfigParser, self).__init__(config_file, config_base_dir='dlk/configures/core/layers/embeddings/')


@config_parser_register('module')
class ModuleConfigParser(BaseConfigParser):
    """docstring for ModuleConfigParser"""
    def __init__(self, config_file):
        super(ModuleConfigParser, self).__init__(config_file, config_base_dir='dlk/configures/core/modules/')

@config_parser_register('processor')
class ProcessorConfigParser(BaseConfigParser):
    """docstring for ProcessorConfigParser"""
    def __init__(self, config_file):
        super(ProcessorConfigParser, self).__init__(config_file, config_base_dir='dlk/configures/data/processors/')


@config_parser_register('subprocessor')
class SubProcessorConfigParser(BaseConfigParser):
    """docstring for SubProcessorConfigParser"""
    def __init__(self, config_file):
        super(SubProcessorConfigParser, self).__init__(config_file, config_base_dir='dlk/configures/data/subprocessors/')


@config_parser_register('postprocessor')
class PostProcessorConfigParser(BaseConfigParser):
    """docstring for PostProcessorConfigParser"""
    def __init__(self, config_file):
        super(PostProcessorConfigParser, self).__init__(config_file, config_base_dir='dlk/configures/data/postprocessors/')




from typing import Dict
from . import callback_register, callback_config_register
from pytorch_lightning.callbacks import LearningRateMonitor


@callback_config_register('lr_monitor')
class LearningRateMonitorCallbackConfig(object):
    """docstring for LearningRateMonitorCallbackConfig
        {
            "_name": "lr_monitor",
            "config": {
                "logging_interval": null, // set to None to log at individual interval according to the interval key of each scheduler. other value : step, epoch
                "log_momentum": true, // log momentum or not
            }
        }
    """
    def __init__(self, config: Dict):
        config = config['config']
        self.logging_interval = config["logging_interval"]
        self.log_momentum = config["log_momentum"]


@callback_register('lr_monitor')
class LearningRateMonitorCallback(object):
    """
    """

    def __init__(self, config: LearningRateMonitorCallbackConfig):
        super().__init__()
        self.config = config

    def __call__(self, rt_config):
        """TODO: Docstring for __call__.

        :rt_config: Dict: TODO
        :returns: TODO

        """
        return LearningRateMonitor(**self.config.__dict__)



import multiprocessing
from gensim.models import Word2Vec
from gensim.models.callbacks import CallbackAny2Vec
import hjson
import argparse
# Data from https://pytorch.org/text/_modules/torchtext/datasets/language_modeling.html
from gensim.models.word2vec_inner import MAX_WORDS_IN_BATCH

def any2unicode(text, encoding='utf8', errors='strict'):
    """Convert `text` (bytestring in given encoding or unicode) to unicode.

    Parameters
    ----------
    text : str
        Input text.
    errors : str, optional
        Error handling behaviour if `text` is a bytestring.
    encoding : str, optional
        Encoding of `text` if it is a bytestring.

    Returns
    -------
    str
        Unicode version of `text`.

    """
    if isinstance(text, str):
        return text
    return str(text, encoding, errors=errors)

class Sentences(object):
    def __init__(self, file_names, max_sentence_length=MAX_WORDS_IN_BATCH):
        self.file_names = file_names
        self.max_sentence_length = max_sentence_length

    def __iter__(self):
        for fname in self.file_names:
            print(f"Training on {fname}")
            for line in open(fname, 'r'):
                line = any2unicode(line).split()
                i = 0
                while i < len(line):
                    yield line[i: i + self.max_sentence_length]
                    i += self.max_sentence_length

class callback(CallbackAny2Vec):
    '''Callback to print loss after each epoch.'''

    def __init__(self):
        self.epoch = 0
        self.loss_to_be_subed = 0

    def on_epoch_end(self, model):
        loss = model.get_latest_training_loss()
        loss_now = loss - self.loss_to_be_subed
        self.loss_to_be_subed = loss
        print('Loss after epoch {}: {}'.format(self.epoch, loss_now))
        self.epoch += 1


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--config",
        type=str,
        default='./word2vec.hjson',
        help="The config path.",
    )
    parser.add_argument(
        "--train_files", type=list, default=['train.txt'], help="train tokenizer files."
    )
    parser.add_argument(
        "--embedding_size",
        type=int,
        default=128,
        help=( "vector size")
    )
    parser.add_argument(
        "--min_count",
        type=int,
        default=10,
        help=( "minist word frequence")
    )
    parser.add_argument(
        "--max_vocab_size",
        type=int,
        default=300000,
        help=( "max vocab size")
    )
    parser.add_argument(
        "--workers",
        type=int,
        default=0,
        help=( "0 means use all cpus")
    )
    parser.add_argument(
        "--window",
        type=int,
        default=6,
        help=( "window size")
    )
    parser.add_argument(
        "--sg",
        type=bool,
        default=0,
        help=( "use skip gram or not. if set to 0 use cbow")
    )
    parser.add_argument(
        "--hs",
        type=bool,
        default=1,
        help=( "If 1, hierarchical softmax will be used for model training.  If 0, and `negative` is non-zero, negative sampling will be used.")
    )
    parser.add_argument(
        "--epochs",
        type=int,
        default=15,
        help=( "training epochs")
    )
    parser.add_argument(
        "--embedding_file",
        type=str,
        default='embedding.txt',
        help=( "embedding file")
    )
    args = parser.parse_args()
    if args.config:
        config_json = hjson.load(open(args.config), object_pairs_hook=dict)
        for key, value in config_json.items():
           setattr(args, key, value)
    if not args.workers:
        args.workers = multiprocessing.cpu_count()
    # print(args)

    print("Start training...")
    lines = Sentences(args.train_files)

    model = Word2Vec(lines, vector_size=args.embedding_size, hs=args.hs, window=args.window, min_count=args.min_count, workers=args.workers, epochs=args.epochs, sg=args.sg, compute_loss=True, max_vocab_size=args.max_vocab_size, callbacks=[callback()])
    print("Saving embedding...")
    model.save(args.model_file)
    model.wv.save_word2vec_format(args.embedding_file, binary=False)



import torch.nn as nn
from dlk.utils.config import BaseConfig
import torch
from typing import Dict, List, Collection
from . import module_register, module_config_register

@module_config_register("conv1d")
class Conv1dConfig(BaseConfig):
    """docstring for Conv1dConfig
    {
        "config": {
            "in_channels": "*@*",
            "out_channels": "*@*",
            "dropout": 0.0,
            "kernel_sizes": [3],
        },
        "_name": "conv1d",
    }
    """
    def __init__(self, config: Dict):
        super(Conv1dConfig, self).__init__(config)
        config = config['config']
        self.kernel_sizes = config['kernel_sizes']
        out_channels = config['out_channels']
        assert all(k % 2 == 1 for k in self.kernel_sizes), 'the kernel sizes must be odd'
        assert out_channels % len(self.kernel_sizes) == 0, 'out channels must be dividable by kernels'
        self.in_channels = config['in_channels']
        self.out_channels = out_channels // len(self.kernel_sizes)
        self.dropout = config['dropout']
        self.post_check(config, used=[
            "in_channels",
            "out_channels",
            "dropout",
            "kernel_sizes",
        ])

@module_register("conv1d")
class Conv1d(nn.Module):
    def __init__(self, config: Conv1dConfig):
        super().__init__()
        convs = []
        for kernel_size in config.kernel_sizes:
            conv = nn.Conv1d(config.in_channels, config.out_channels, kernel_size,
                             padding=(kernel_size - 1) // 2)
            convs.append(nn.Sequential(conv, nn.GELU()))
        self.convs = nn.ModuleList(convs)
        self.dropout = nn.Dropout(p=float(config.dropout))

    def forward(self, x):
        return self.dropout(torch.cat([conv(x) for conv in self.convs], dim=-1))



from . import model_register, model_config_register
from typing import Dict, List
from dlk.utils.config import BaseConfig, ConfigTool
from dlk.core.base_module import BaseModel
import torch
from dlk.core.layers.embeddings import embedding_config_register, embedding_register
from dlk.core.initmethods import initmethod_config_register, initmethod_register
from dlk.core.layers.encoders import encoder_config_register, encoder_register
from dlk.core.layers.decoders import decoder_config_register, decoder_register


@model_config_register('basic')
class BasicModelConfig(BaseConfig):
    """docstring for BasicIModelConfig
    {
        embedding: {
            _base: "static"
            config: {
                embedding_file: "*@*", //the embedding file, must be saved as numpy array by pickle
                embedding_dim: "*@*",
                //if the embedding_file is a dict, you should provide the dict trace to embedding
                embedding_trace: ".", //default the file itself is the embedding
                /*embedding_trace: "embedding", //this means the <embedding = pickle.load(embedding_file)["embedding"]>*/
                /*embedding_trace: "meta.embedding", //this means the <embedding = pickle.load(embedding_file)['meta']["embedding"]>*/
                freeze: false, // is freeze
                dropout: 0, //dropout rate
                output_map: {},
            },
        },
        decoder: {
            _base: "linear",
            config: {
                input_size: "*@*",
                output_size: "*@*",
                pool: null,
                dropout: "*@*", //the decoder output no need dropout
                output_map: {}
            },
        },
        encoder: {
            _base: "lstm",
            config: {
                output_map: {},
                hidden_size: "*@*",
                input_size: *@*,
                output_size: "*@*",
                num_layers: 1,
                dropout: "*@*", // dropout between layers
            },
        },
        "initmethod": {
            "_base": "range_norm"
        },
        "config": {
            "embedding_dim": "*@*",
            "dropout": "*@*",
            "embedding_file": "*@*",
            "embedding_trace": "token_embedding",
        },
        _link: {
            "config.embedding_dim": ["embedding.config.embedding_dim",
                                     "encoder.config.input_size",
                                     "encoder.config.output_size",
                                     "encoder.config.hidden_size",
                                     "decoder.config.output_size",
                                     "decoder.config.input_size"
                                    ],
            "config.dropout": ["encoder.config.dropout", "decoder.config.dropout", "embedding.config.dropout"],
            "config.embedding_file": ['embedding.config.embedding_file'],
            "config.embedding_trace": ['embedding.config.embedding_trace']
        }
        _name: "basic"
    }
    """
    def __init__(self, config):
        super(BasicModelConfig, self).__init__(config)

        self.embedding, self.embedding_config = self.get_embedding(config["embedding"])
        self.encoder, self.encoder_config = self.get_encoder(config["encoder"])
        self.decoder, self.decoder_config = self.get_decoder(config["decoder"])
        self.init_method, self.init_method_config = self.get_init_method(config["initmethod"])
        self.config = config['config']

    def get_embedding(self, config):
        """get embedding config and embedding module

        :config: TODO
        :returns: TODO

        """
        return ConfigTool.get_leaf_module(embedding_register, embedding_config_register, "embedding", config)

    def get_init_method(self, config):
        """get embedding config and embedding module

        :config: TODO
        :returns: TODO

        """
        return ConfigTool.get_leaf_module(initmethod_register, initmethod_config_register, "init method", config)

    def get_encoder(self, config):
        """get encoder config and encoder module

        :config: TODO
        :returns: TODO

        """
        return ConfigTool.get_leaf_module(encoder_register, encoder_config_register, "encoder", config)

    def get_decoder(self, config):
        """get decoder config and decoder module

        :config: TODO
        :returns: TODO
        """
        return ConfigTool.get_leaf_module(decoder_register, decoder_config_register, "decoder", config)


@model_register('basic')
class BasicModel(BaseModel):
    """
    Sequence labeling model
    """

    def __init__(self, config: BasicModelConfig, checkpoint):
        super().__init__()

        self.embedding = config.embedding(config.embedding_config)
        self.encoder = config.encoder(config.encoder_config)
        self.decoder = config.decoder(config.decoder_config)

        if not checkpoint:
            init_method = config.init_method(config.init_method_config)
            self.embedding.init_weight(init_method)
            self.encoder.init_weight(init_method)
            self.decoder.init_weight(init_method)

        self.config = config.config
        self._provided_keys = self.config.get("provided_keys", [])

    def provide_keys(self)->List[str]:
        """TODO: should provide_keys in model?
        """
        return self.decoder.provided_keys()

    def check_keys_are_provided(self, provide: List[str]=[])->None:
        """TODO: should check keys in model?
        """
        self._provided_keys = self._provided_keys + provide
        self.embedding.check_keys_are_provided(self._provided_keys)
        self.encoder.check_keys_are_provided(self.embedding.provide_keys())
        self.decoder.check_keys_are_provided(self.encoder.provide_keys())

    def forward(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """forward
        :inputs: Dict[str: torch.Tensor], one mini-batch inputs
        :returns: Dict[str: torch.Tensor], one mini-batch outputs
        """
        embedding_outputs = self.embedding(inputs)
        encode_outputs = self.encoder(embedding_outputs)
        decode_outputs = self.decoder(encode_outputs)
        return decode_outputs

    def predict_step(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """predict
        :inputs: Dict[str: torch.Tensor], one mini-batch inputs
        :returns: Dict[str: torch.Tensor], one mini-batch outputs
        """
        embedding_outputs = self.embedding.predict_step(inputs)
        encode_outputs = self.encoder.predict_step(embedding_outputs)
        decode_outputs = self.decoder.predict_step(encode_outputs)
        return decode_outputs

    def training_step(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """training
        :inputs: Dict[str: torch.Tensor], one mini-batch inputs
        :returns: Dict[str: torch.Tensor], one mini-batch outputs
        """
        embedding_outputs = self.embedding.training_step(inputs)
        encode_outputs = self.encoder.training_step(embedding_outputs)
        decode_outputs = self.decoder.training_step(encode_outputs)
        return decode_outputs

    def validation_step(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """valid
        :inputs: Dict[str: torch.Tensor], one mini-batch inputs
        :returns: Dict[str: torch.Tensor], one mini-batch outputs
        """
        embedding_outputs = self.embedding.validation_step(inputs)
        encode_outputs = self.encoder.validation_step(embedding_outputs)
        decode_outputs = self.decoder.validation_step(encode_outputs)
        return decode_outputs

    def test_step(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """valid
        :inputs: Dict[str: torch.Tensor], one mini-batch inputs
        :returns: Dict[str: torch.Tensor], one mini-batch outputs
        """
        embedding_outputs = self.embedding.test_step(inputs)
        encode_outputs = self.encoder.test_step(embedding_outputs)
        decode_outputs = self.decoder.test_step(encode_outputs)
        return decode_outputs



import torch.nn as nn
import torch
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence
from typing import Dict
from dlk.utils.config import BaseConfig
from . import module_register, module_config_register
from dlk.utils.logger import Logger

logger = Logger.get_logger()

@module_config_register("lstm")
class LSTMConfig(BaseConfig):
    """docstring for LSTMConfig
    {
        "config": {
            "bidirectional": true,
            "output_size": 200, //the output is 2*hidden_size if use
            "input_size": 200,
            "num_layers": 1,
            "dropout": 0.1, // dropout between layers
            "dropout_last": true, //dropout the last layer output or not
        },
        "_name": "lstm",
    }
    """

    def __init__(self, config: Dict):
        super(LSTMConfig, self).__init__(config)
        config = config['config']
        self.num_layers = config['num_layers']
        self.bidirectional= config['bidirectional']
        self.input_size = config['input_size']
        self.output_size = config['output_size']
        self.hidden_size = self.output_size
        if self.bidirectional:
            assert self.output_size % 2 == 0
            self.hidden_size = self.output_size // 2
        self.dropout = config['dropout']
        self.dropout_last = config['dropout_last']
        self.post_check(config, used=[
            "bidirectional",
            "output_size",
            "input_size",
            "num_layers",
            "dropout",
            "dropout_last",
        ])


@module_register("lstm")
class LSTM(nn.Module):
    def __init__(self, config: LSTMConfig):
        super(LSTM, self).__init__()

        if config.num_layers <= 1:
            inlstm_dropout = 0
        else:
            inlstm_dropout = config.dropout
        self.lstm = nn.LSTM(input_size=config.input_size, hidden_size=config.hidden_size, num_layers=config.num_layers, batch_first=True, bidirectional=config.bidirectional, dropout=inlstm_dropout)
        self.dropout_last = nn.Dropout(p=float(config.dropout) if config.dropout_last else 0)

    def forward(self, input: torch.Tensor, mask: torch.Tensor)->torch.Tensor:
        """
        """
        max_seq_len = input.size(1)
        seq_lens = mask.sum(1).cpu()
        pack_seq_rep = pack_padded_sequence(input=input, lengths=seq_lens, batch_first=True, enforce_sorted=False)
        pack_seq_rep = self.lstm(pack_seq_rep)[0]
        output, _ = pad_packed_sequence(sequence=pack_seq_rep, batch_first=True, total_length=max_seq_len)

        return self.dropout_last(output)



from dlk.utils.logger import setting_logger
import pytorch_lightning as pl
# setting_logger("./benchmark.log")
# setting_logger("./conll_norm_lstm_crf.log")
setting_logger("./conll_norm_char_lstm_crf.log")
from dlk.train import Train
import json

pl.seed_everything(88)

# trainer = Train('./examples/sequence_labeling/benchmark/pretrained/first_piece_lstm_crf_main.hjson')
trainer = Train('./examples/sequence_labeling/conll2003/norm_char_lstm_crf/main.hjson')
# trainer = Train('./examples/sequence_labeling/conll2003/norm_lstm_crf/main.hjson')

trainer.run()



import torch.nn as nn
from . import initmethod_register, initmethod_config_register
from typing import Dict, List
from dlk.utils.logger import Logger
from dlk.utils.config import BaseConfig
import numpy as np
import torch

logger = Logger.get_logger()

# TODO:
@initmethod_config_register('default')
class RangeNormInitConfig(BaseConfig):
    """
        {
            "_name": "default",
            "config": {
            }
        }
    """
    def __init__(self, config):
        super(RangeNormInitConfig, self).__init__(config)
        self.post_check(config['config'])

@initmethod_register('default')
class RangeNormInit(object):
    """
    """

    def __init__(self, config: RangeNormInitConfig):
        super().__init__()

    def __call__(self, module):
        """Initialize the weights"""
        if isinstance(module, nn.Linear):
            # use the default kaiming init method
            torch.nn.init.xavier_uniform_(module.weight)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            torch.nn.init.xavier_uniform_(module.weight)
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)
        elif isinstance(module, nn.Conv1d):
            torch.nn.init.kaiming_uniform_(module.weight)
        elif isinstance(module, nn.Conv2d):
            torch.nn.init.kaiming_uniform_(module.weight)
        elif isinstance(module, nn.Conv3d):
            torch.nn.init.kaiming_uniform_(module.weight)
        elif isinstance(module, nn.LSTM):
            self.init_lstm(module)
        elif isinstance(module, nn.ModuleList) or isinstance(module, nn.Sequential):
            for sub_module in module:
                self(sub_module)
        elif isinstance(module, nn.ModuleDict):
            for sub_module in module:
                self(module[sub_module])
        else:
            logger.info(f"{module} is not initialization.")

    def init_lstm(self, lstm):
        """
        Initialize lstm
        """
        # nn.init.xavier_uniform_(lstm.weight_hh_l0)
        # nn.init.xavier_uniform_(lstm.weight_hh_l0_reverse)
        # nn.init.xavier_uniform_(lstm.weight_ih_l0)
        # nn.init.xavier_uniform_(lstm.weight_ih_l0_reverse)
        # lstm.bias_hh_l0.data.fill_(0)
        # lstm.bias_hh_l0_reverse.data.fill_(0)
        # lstm.bias_ih_l0.data.fill_(0)
        # lstm.bias_ih_l0_reverse.data.fill_(0)
        # # Init forget gates to 1
        # for names in lstm._all_weights:
            # for name in filter(lambda n: 'bias' in n, names):
                # bias = getattr(lstm, name)
                # n = bias.size(0)
                # start, end = n // 4, n // 2
                # bias.data[start:end].fill_(1.)

        # Another init method
        for ind in range(0, lstm.num_layers):
            weight = eval('lstm.weight_ih_l' + str(ind))
            bias = np.sqrt(6.0 / (weight.size(0) / 4 + weight.size(1)))
            nn.init.uniform_(weight, -bias, bias)
            weight = eval('lstm.weight_hh_l' + str(ind))
            bias = np.sqrt(6.0 / (weight.size(0) / 4 + weight.size(1)))
            nn.init.uniform_(weight, -bias, bias)
        if lstm.bidirectional:
            for ind in range(0, lstm.num_layers):
                weight = eval('lstm.weight_ih_l' + str(ind) + '_reverse')
                bias = np.sqrt(6.0 / (weight.size(0) / 4 + weight.size(1)))
                nn.init.uniform_(weight, -bias, bias)
                weight = eval('lstm.weight_hh_l' + str(ind) + '_reverse')
                bias = np.sqrt(6.0 / (weight.size(0) / 4 + weight.size(1)))
                nn.init.uniform_(weight, -bias, bias)

        if lstm.bias:
            for ind in range(0, lstm.num_layers):
                weight = eval('lstm.bias_ih_l' + str(ind))
                weight.data.zero_()
                weight.data[lstm.hidden_size: 2 * lstm.hidden_size] = 1
                weight = eval('lstm.bias_hh_l' + str(ind))
                weight.data.zero_()
                weight.data[lstm.hidden_size: 2 * lstm.hidden_size] = 1
            if lstm.bidirectional:
                for ind in range(0, lstm.num_layers):
                    weight = eval('lstm.bias_ih_l' + str(ind) + '_reverse')
                    weight.data.zero_()
                    weight.data[lstm.hidden_size: 2 * lstm.hidden_size] = 1
                    weight = eval('lstm.bias_hh_l' + str(ind) + '_reverse')
                    weight.data.zero_()
                    weight.data[lstm.hidden_size: 2 * lstm.hidden_size] = 1



import pandas as pd
from torch.utils.data import DataLoader, random_split, Dataset
from typing import Dict, List, Union, Any
from dlk.utils.config import BaseConfig, ConfigTool
from dlk.data.datamodules import datamodule_config_register, datamodule_register, IBaseDataModule, collate_register
from dlk.utils.logger import Logger
# from pytorch_lightning import LightningDataModule
from torch.nn.utils.rnn import pad_sequence
import torch
import copy
logger = Logger.get_logger()

@datamodule_config_register('basic')
class BasicDatamoduleConfig(BaseConfig):
    """docstring for BasicDatamoduleConfig
       {
           "_name": "basic",
           "config": {
               "pin_memory": None,
               "collate_fn": "default",
               "shuffle": {
                   "train": true,
                   "predict": false,
                   "valid": false,
                   "test": false,
                   "online": false
               },
               "key_type_pairs": {
                    'input_ids': 'int',
                    'label_ids': 'long',
                    'type_ids': 'long',
                },
               "gen_mask": {
                    'input_ids': 'attention_mask',
                },
               "key_padding_pairs": { //default all 0
                    'input_ids': 0,
                },
               "train_batch_size": 32,
               "predict_batch_size": 32, //predict、test batch_size is equals to valid_batch_size
               "online_batch_size": 1,
           }
       },
    """
    def __init__(self, config):
        super(BasicDatamoduleConfig, self).__init__(config)
        config = config['config']
        self.key_type_pairs = config.get('key_type_pairs', {})

        if "_index" in self.key_type_pairs:
            logger.warning(f"'_index' is a preserved key, we will use this to indicate the index of item, if you ignore this warning, we will ignore the origin '_index' data.")
        self.key_padding_pairs = config.get('key_padding_pairs', {})
        self.gen_mask = config.get("gen_mask", {})
        self.collate_fn = config.get('collate_fn', 'default')
        self.pin_memory = config.get('pin_memory', False)
        if self.pin_memory is None:
            self.pin_memory = torch.cuda.is_available()
        self.shuffle = config.get('shuffle', {
            "train": True,
            "predict": False,
            "valid": False,
            "test": False,
            "online": False
        })
        self.train_batch_size = config.get('train_batch_size', 32)
        self.test_batch_size = config.get('predict_batch_size', 32)
        self.valid_batch_size = config.get('predict_batch_size', 32)
        self.predict_batch_size = config.get('predict_batch_size', 32)
        self.online_batch_size = config.get('online_batch_size', 1)
        self.post_check(config, used=[
               "pin_memory",
               "collate_fn",
               "shuffle",
               "key_type_pairs",
               "gen_mask",
               "key_padding_pairs",
               "train_batch_size",
               "predict_batch_size",
               "online_batch_size",
        ])


class BasicDataset(Dataset):
    def __init__(self, key_type_pairs: Dict[str, str], data:pd.DataFrame):
        self.data = data
        self.type_map = {"float": torch.float, "int": torch.int, 'bool': torch.bool, "long": torch.long}

        self.key_type_pairs = key_type_pairs

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        one_ins = {}
        for key, key_type in self.key_type_pairs.items():
            one_ins[key] = torch.tensor(self.data.iloc[idx][key], dtype=self.type_map[key_type])
        one_ins['_index'] = torch.tensor(idx,  dtype=torch.long)
        return one_ins


@datamodule_register("basic")
class BasicDatamodule(IBaseDataModule):
    def __init__(self, config: BasicDatamoduleConfig, data: Dict[str, Any]):
        super().__init__()

        self.config = config
        self.train_data = None
        self.valid_data = None
        self.test_data = None
        self.predict_data = None
        if 'train' in data:
            self.train_data = BasicDataset(self.real_key_type_pairs(config.key_type_pairs, data, 'train'), data['train'])
        if "test" in data:
            self.test_data = BasicDataset(self.real_key_type_pairs(config.key_type_pairs, data, 'test'), data['test'])
        if "valid" in data:
            self.valid_data = BasicDataset(self.real_key_type_pairs(config.key_type_pairs, data, 'valid'), data['valid'])
        if "predict" in data:
            self.predict_data = BasicDataset(self.real_key_type_pairs(config.key_type_pairs, data, 'predict'), data['predict'])
        self.collate_fn = collate_register.get(config.collate_fn)(key_padding_pairs=config.key_padding_pairs, gen_mask=config.gen_mask)

    def real_key_type_pairs(self, key_type_pairs: Dict, data: Dict, field: str):
        copy_key_type_pairs = copy.deepcopy(key_type_pairs)
        has_key = set(data[field].columns)
        remove = set()
        for key in copy_key_type_pairs:
            if key not in has_key:
                remove.add(key)
        if remove:
            logger.warning(f"There are not '{', '.join(remove)}' in data field {field}.")
        for key in remove:
            copy_key_type_pairs.pop(key)
        return copy_key_type_pairs

    def train_dataloader(self):
        if not self.train_data:
            return None
        return DataLoader(self.train_data, batch_size=self.config.train_batch_size, collate_fn=self.collate_fn, pin_memory=self.config.pin_memory, shuffle=self.config.shuffle.get('train', True))

    def predict_dataloader(self):
        """
        """
        if not self.predict_data:
            return None
        return DataLoader(self.predict_data, batch_size=self.config.predict_batch_size, collate_fn=self.collate_fn, pin_memory=self.config.pin_memory, shuffle=self.config.shuffle.get('predict', False))

    def val_dataloader(self):
        if not self.valid_data:
            return None
        return DataLoader(self.valid_data, batch_size=self.config.valid_batch_size, collate_fn=self.collate_fn, pin_memory=self.config.pin_memory, shuffle=self.config.shuffle.get('valid', False))

    def test_dataloader(self):
        if not self.test_data:
            return None
        return DataLoader(self.test_data, batch_size=self.config.test_batch_size, collate_fn=self.collate_fn, pin_memory=self.config.pin_memory, shuffle=self.config.shuffle.get('test', False))

    def online_dataloader(self):
        # return DataLoader(self.mnist_test, batch_size=self.batch_size)
        return self.collate_fn



import torch.nn as nn
import torch
from typing import Dict, List
from dlk.utils.config import BaseConfig
from . import module_register, module_config_register


@module_config_register("logits_gather")
class LogitsGatherConfig(BaseConfig):
    """docstring for LogitsGatherConfig
    {
        "config": {
            "gather_layer": {
                "0": {
                    "map": "3", // the 0th layer not do scale output to "gather_logits_3", "gather_logits_" is the output name prefix, the "3" is map name
                    "scale": {} //don't scale
                },
                "1": {
                    "map": "4",  // the 1th layer scale output dim from 1024 to 200 and the output named "gather_logits_3"
                    "scale": {"1024":"200"},
                }
            },
            "prefix": "gather_logits_",
        },
        _name: "logits_gather",
    }
    """
    def __init__(self, config: Dict):
        if '_name' not in config:
            config['_name'] = 'logits_gather'
        super(LogitsGatherConfig, self).__init__(config)
        config = config.get('config', {})
        self.gather_layer = config.get('gather_layer', {})
        self.prefix = config.get("prefix", '')
        self.post_check(config, used=[
            "gather_layer",
            "prefix"
        ])


@module_register("logits_gather")
class LogitsGather(nn.Module):
    def __init__(self, config: LogitsGatherConfig):
        super(LogitsGather, self).__init__()
        gather_layer_num = len(config.gather_layer)
        self.layers_scale = nn.ModuleDict()
        self.layer_map: Dict[str, str] = {}
        self.prefix = config.prefix
        for layer, layer_config in config.gather_layer.items():
            self.layer_map[str(layer)] = str(layer_config['map'])
            if layer_config.get("scale", {}):
                scale = layer_config['scale']
                assert len(scale) == 1
                for from_dim, to_dim in scale.items:
                    self.layers_scale[str(layer)] = nn.Linear(int(from_dim), int(to_dim))

    def forward(self, input: List[torch.Tensor])->Dict[str, torch.Tensor]:
        """
        """
        result: Dict[str, torch.Tensor] = {"a": torch.tensor([0])}
        if not self.layer_map:
            return result
        for layer, layer_suffix in self.layer_map.items():
            result[self.prefix+layer_suffix] = self.layers_scale[layer](input[int(layer)])

        return result



from dlk.utils.parser import config_parser_register
import json
import hjson
from typing import Dict, Union, Any
from dlk.data.processors import processor_config_register, processor_register


class Processor(object):
    """docstring for Processor"""
    def __init__(self, config: Union[str, Dict]):
        super(Processor, self).__init__()
        self.config = config
        if not isinstance(config, dict):
            config = hjson.load(open(config), object_pairs_hook=dict)
            config = config_parser_register.get("processor")(config).parser_with_check()
            assert len(config) == 1, f"Currently we didn't support search for Processor, if you require this feature please create an issue to describe the reason details."
            self.config = config[0]
        
    def fit(self, data: Dict[str, Any], stage='train'):
        """TODO: Docstring for fit.
        :returns: TODO

        """
        processor = processor_register.get(self.config.get('_name'))(stage=stage, config=processor_config_register.get(self.config.get('_name'))(stage=stage, config=self.config))
        return processor.process(data)



from typing import Dict
import torch.nn as nn
import torch.optim as optim
from dlk.utils.config import BaseConfig
from . import optimizer_register, optimizer_config_register, BaseOptimizer


@optimizer_config_register("adamw")
class AdamWOptimizerConfig(BaseConfig):
    """docstring for LinearConfig
    {
        "config": {
            "lr": 5e-5,
            "betas": [0.9, 0.999],
            "eps": 1e-6,
            "weight_decay": 1e-2,
            "optimizer_special_groups": {
                "order": ['decoder', 'bias'], // the group order, if the para is in decoder & is in bias, set to decoder
                "bias": {
                    "config": {
                        "weight_decay": 0
                    },
                    "pattern": ["bias",  "LayerNorm.bias", "LayerNorm.weight"]
                },
                "decoder": {
                    "config": {
                        "lr": 1e-3
                    },
                    "pattern": ["decoder"]
                },
            }
            "name": "default" // default group name
        },
        "_name": "adamw",
    }
    """
    def __init__(self, config: Dict):
        super(AdamWOptimizerConfig, self).__init__(config)
        self.config = config['config']
        self.post_check(self.config, used=[
            "lr",
            "betas",
            "eps",
            "weight_decay",
            "optimizer_special_groups",
            "name",
        ])


@optimizer_register("adamw")
class AdamWOptimizer(BaseOptimizer):
    def __init__(self, model: nn.Module, config: AdamWOptimizerConfig):
        super(AdamWOptimizer, self).__init__()
        self.config = config.config
        self.model = model
        self.optimizer = optim.AdamW

    def get_optimizer(self):
        """TODO: Docstring for __call__.
        :returns: TODO

        """

        return self.init_optimizer(optim.AdamW, self.model, self.config)



"""schedulers"""
import importlib
import os
from dlk.utils.register import Register
from torch.optim import Optimizer
from torch.optim.lr_scheduler import LambdaLR
import math


scheduler_config_register = Register("Schedule config register.")
scheduler_register = Register("Schedule register.")


class BaseScheduler(object):
    """docstring for BaseSchedule"""

    def get_scheduler(self):
        """TODO: Docstring for get_scheduler.
        :returns: TODO

        """
        raise NotImplementedError

    def __call__(self):
        """TODO: Docstring for __call__.

        :arg1: TODO
        :returns: TODO

        """
        return self.get_scheduler()


def import_schedulers(schedulers_dir, namespace):
    for file in os.listdir(schedulers_dir):
        path = os.path.join(schedulers_dir, file)
        if (
            not file.startswith("_")
            and not file.startswith(".")
            and (file.endswith(".py") or os.path.isdir(path))
        ):
            scheduler_name = file[: file.find(".py")] if file.endswith(".py") else file
            importlib.import_module(namespace + "." + scheduler_name)


# automatically import any Python files in the schedulers directory
schedulers_dir = os.path.dirname(__file__)
import_schedulers(schedulers_dir, "dlk.core.schedulers")



import hjson
import pandas as pd
from typing import Union, Dict
from dlk.utils.parser import BaseConfigParser
from dlk.data.processors import IProcessor, processor_config_register, processor_register
from dlk.data.subprocessors import subprocessor_config_register, subprocessor_register
from dlk.utils.config import BaseConfig
from dlk.utils.logger import Logger

logger = Logger.get_logger()

@processor_config_register('basic')
class BasicProcessorConfig(BaseConfig):
    """docstring for BasicProcessorConfig
    config e.g.
    {
        // input should be {"train": train, "valid": valid, ...}, train/valid/test/predict/online etc, should be dataframe and must have a column named "sentence"
        "_name": "basic@test_text_cls",
        "config": {
            "feed_order": ["load", "tokenizer", "token_gather", "label_to_id", "token_embedding", "save"]
        },
        "subprocessor@load": {
            "_base": "load",
            "config":{
                "base_dir": "."
                "predict":{
                    "meta": "./meta.pkl",
                },
                "online": [
                    "predict", //base predict
                    {   // special config, update predict, is this case, the config is null, means use all config from "predict", when this is empty dict, you can only set the value to a str "predict", they will get the same result
                    }
                ]
            }
        },
        "subprocessor@save": {
            "_base": "save",
            "config":{
                "base_dir": "."
                "train":{
                    "processed": "processed_data.pkl", // all data
                    "meta": {
                        "meta.pkl": ['label_vocab'] //only for next time use
                    }
                },
                "predict": {
                    "processed": "processed_data.pkl",
                }
            }
        },
        "subprocessor@tokenizer":{
            "_base": "fast_tokenizer",
            "config": {
                "train": {
                    "config_path": "*@*",
                    "prefix": ""
                    "data_type": "single", // single or pair, if not provide, will calc by len(process_data)
                    "process_data": [
                        ["sentence", { "is_pretokenized": false}],
                    ],
                    "post_processor": "default"
                    "filed_map": { // this is the default value, you can provide other name
                        "ids": "input_ids",
                    }, // the tokenizer output(the key) map to the value
                },
                "predict": "train",
                "online": "train"
            }
        },
        "subprocessor@token_gather":{
            "_base": "token_gather",
            "config": {
                "train": { // only train stage using
                    "data_set": {      // for different stage, this processor will process different part of data
                        "train": ["train", "valid"]
                    },
                    "gather_columns": ["label"], //List of columns. Every cell must be sigle token or list of tokens or set of tokens
                    "deliver": "label_vocab", // output Vocabulary object (the Vocabulary of labels) name.
                }
            }
        },
        "subprocessor@label_to_id":{
            "_base": "token2id",
            "config": {
                "train":{ //train、predict、online stage config,  using '&' split all stages
                    "data_pair": {
                        "label": "label_id"
                    },
                    "data_set": {                   // for different stage, this processor will process different part of data
                        "train": ['train', 'valid', 'test'],
                        "predict": ['predict'],
                        "online": ['online']
                    },
                    "vocab": "label_vocab", // usually provided by the "token_gather" module
                }, //3
                "predict": "train",
                "online": "train",
            }
        },
        "subprocessor@token_embedding": {
            "_base": "token_embedding",
            "config":{
                "train": { // only train stage using
                    "embedding_file": "*@*",
                    "tokenizer": "*@*", //List of columns. Every cell must be sigle token or list of tokens or set of tokens
                    "deliver": "token_embedding", // output Vocabulary object (the Vocabulary of labels) name.
                    "embedding_size": 200,
                }
            }
        },
    }
    """

    def __init__(self, stage, config: Dict):
        super(BasicProcessorConfig, self).__init__(config)
        if isinstance(config, str):
            config = hjson.load(open(config), object_pairs_hook=dict)
        self.feed_order = config["config"]['feed_order']
        self.subprocessors = config
        self.stage = stage
        # self.post_check(config['config'], used=['feed_order'])


@processor_register('basic')
class BasicProcessor(IProcessor):
    """docstring for DataSet"""
    def __init__(self, stage: str, config: BasicProcessorConfig):
        super(BasicProcessor, self).__init__()
        self._name = "basic"
        self.stage = stage
        self.feed_order = config.feed_order
        assert len(self.feed_order) > 0

        self.subprocessors = {}
        for name in self.feed_order:
            subprocessor_config_dict = config.subprocessors[f'subprocessor@{name}']
            logger.info(f"Init '{name}' ....")
            subprocessor_name = subprocessor_config_dict["_name"]
            subprocessor_config = subprocessor_config_register.get(subprocessor_name)(stage=self.stage, config=subprocessor_config_dict)
            subprocessor = subprocessor_register.get(subprocessor_name)(stage=self.stage, config=subprocessor_config)
            self.subprocessors[name] = subprocessor

    def process(self, data: Dict)->Dict:
        """TODO: Docstring for process.

        :data: TODO
        :returns: TODO
        """
        logger.info(f"Start Data Processing....")
        for name in self.feed_order:
            if self.stage != 'online':
                logger.info(f"Processing on '{name}' ....")
            data = self.subprocessors[name].process(data)
        logger.info(f"Data Processed.")

        return data




from typing import Dict
import torch.nn as nn
from . import loss_register, loss_config_register
from dlk.core.base_module import BaseModuleConfig
import torch.nn as nn


@loss_config_register("identity")
class IdentityLossConfig(BaseModuleConfig):
    """docstring for IdentityLossConfig
    {
        config: {
            "schedule": [1],
            "scale": [1], # scale the loss for every schedule
            // "schedule": [0.3, 1.0], # can be a list or str
            // "scale": "[0.5, 1]",
            "loss": "loss", // the real loss from result['loss']
        },
        _name: "identity",
    }
    """
    def __init__(self, config: Dict):
        super(IdentityLossConfig, self).__init__(config)
        config = config['config']

        self.scale = config['scale']
        self.schedule = config['schedule']
        self.loss = config['loss']

        if isinstance(self.scale, str):
            self.scale = eval(self.scale)
        if isinstance(self.schedule, str):
            self.schedule = eval(self.schedule)

        if not isinstance(self.scale, list):
            assert isinstance(float(self.scale), float)
            self.scale = [self.scale]
        if not isinstance(self.schedule, list):
            assert isinstance(float(self.schedule), float)
            self.schedule = [self.schedule]
        assert len(self.schedule) == len(self.scale)
        assert self.schedule[-1] - 1 < 0.00001
        self.post_check(config, used=[
            "loss",
            "schedule",
            "scale",
        ])

@loss_register("identity")
class IdentityLoss(object):
    def __init__(self, config: IdentityLossConfig):
        super(IdentityLoss, self).__init__()
        self.config = config

    def update_config(self, rt_config):
        """TODO: Docstring for update_config.
        :rt_config: TODO
         {
             "total_steps": self.num_training_steps,
             "total_epochs": self.num_training_epochs
         }
        :returns: TODO

        """
        self.current_stage = 0
        self.config.schedule = [rt_config['total_steps']*i for i in self.config.schedule]

    def calc(self, result, inputs, rt_config):
        """TODO: Docstring for get_loss.
        :returns: TODO
        """
        if rt_config['current_step']>self.config.schedule[self.current_stage]:
            self.current_stage += 1
        scale = self.config.scale[self.current_stage]
        loss = result[self.config.loss] * scale
        return loss

    def __call__(self, result, inputs, rt_config):
        """TODO: Docstring for __call__.
        :returns: TODO
         rt_config={
             "current_step": self.global_step,
             "current_epoch": self.current_epoch,
             "total_steps": self.num_training_steps,
             "total_epochs": self.num_training_epochs
         }),
        """
        return self.calc(result, inputs, rt_config)
